import argparse
import logging

from itertools import chain, combinations

from .constants import DEFAULT_RATES
from .types import Scale, match_key, match_mode
from .constants import NOTE_DURATIONS
from .config import Configuration
from .parser import parse_config_for_generate
from .commands import generate, mutate, continue_sequence


def powerset(iterable):
    return list(map(lambda x: set(x), chain.from_iterable(combinations(iterable, r) for r in range(len(iterable) + 1))))


class CustomParser(argparse.ArgumentDefaultsHelpFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(
            kwargs.get('prog'),
            indent_increment=2,
            max_help_position=100,
            width=None)

    def _format_action(self, action):
        result = super()._format_action(action)
        if isinstance(action, argparse._SubParsersAction):
            # fix indentation on first line
            return "%*s%s" % (self._current_indent, "", result.lstrip())
        return result

    def _format_action_invocation(self, action):
        if isinstance(action, argparse._SubParsersAction):
            # remove metavar and help line
            return ""
        return super()._format_action_invocation(action)

    def _iter_indented_subactions(self, action):
        if isinstance(action, argparse._SubParsersAction):
            try:
                get_subactions = action._get_subactions
            except AttributeError:
                pass
            else:
                # remove indentation
                yield from get_subactions()
        else:
            yield from super()._iter_indented_subactions(action)


def print_cli():
    parser = argparse.ArgumentParser(prog="midi-generator",
                                     description='Command line interface for b2bAI.',
                                     usage='%(prog)s [OPTIONS] COMMAND',
                                     epilog="Run 'midi-generator COMMAND --help' for more information on a command.",
                                     formatter_class=CustomParser)

    parser.add_argument('-l', '--logging-level',
                        choices=[logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        type=int,
                        help='set logging level',
                        default=logging.INFO)

    subparsers = parser.add_subparsers(title='commands', dest='command')

    generate_seq = subparsers.add_parser('generate',
                                         description='Generate a MIDI file from a given configuration.',
                                         help='generate a MIDI file',
                                         usage='%(prog)s [OPTIONS] generate {FILE|OPTIONS}',
                                         formatter_class=CustomParser)

    generate_seq.add_argument('-c', '--config',
                              type=argparse.FileType('r'),
                              help='the configuration file',
                              metavar='File')

    generate_seq.add_argument('--bars',
                              type=int,
                              metavar='{4|8}',
                              help='the number of bars to generate',
                              choices=[4, 8],
                              default=4)

    generate_seq.add_argument('--rates',
                              type=set[float],
                              metavar='set[float]',
                              help='the note lengths used to generate the MIDI file',
                              default=set(DEFAULT_RATES))

    generate_seq.add_argument('--scale',
                              type=str,
                              metavar='Scale',
                              help='the scale in which to generate the MIDI file',
                              default='C minor')

    generate_seq.add_argument('--density',
                              type=float,
                              metavar='float',
                              help='the density of notes in the generated MIDI file',
                              default=0.8)

    generate_seq.add_argument('--syncopation',
                              type=float,
                              metavar='float',
                              help='the level of syncopation in the generated MIDI file',
                              default=0.1)

    mutate_seq = subparsers.add_parser('mutate',
                                       description='Mutate a given MIDI file.',
                                       usage='%(prog)s [OPTIONS] generate FILE [OPTIONS]',
                                       help='mutate a MIDI file.',
                                       formatter_class=CustomParser)

    mutate_seq.add_argument('file',
                            type=str,
                            help='file containing sequence to mutate')

    mutate_seq.add_argument('--probability',
                            type=float,
                            metavar='float',
                            help='the probability of mutating a note',
                            default=0.1)

    continue_seq = subparsers.add_parser('continue',
                                         description='Continue a given MIDI sequence.',
                                         usage='%(prog)s [OPTIONS] generate FILE [OPTIONS]',
                                         help='continue a MIDI sequence.',
                                         formatter_class=CustomParser)

    continue_seq.add_argument('file',
                              type=argparse.FileType('r'),
                              help='file containing sequence to continue')

    args = parser.parse_args()

    logging.basicConfig(level=args.logging_level)
    logging.debug(' arguments given: %s\n', args)

    match args.command:
        case 'generate':
            if args.config is None:
                logging.info(' no configuration file given, using default configuration\n')

                if args.rates not in powerset(NOTE_DURATIONS):
                    logging.warning(' rates not in powerset of note durations, using default rates\n')
                    args.rates = DEFAULT_RATES
                else:
                    args.rates = list(args.rates)

                key = match_key(args.scale.split(' ')[0].lower())
                mode = match_mode(args.scale.split(' ')[1].lower())

                config = Configuration(
                    bars=args.bars,
                    rate=args.rates,
                    scale=Scale(key, mode),
                    density=args.density,
                    syncopation=args.syncopation,
                )
            else:
                logging.info(' using configuration file %s', args.config)
                config = parse_config_for_generate(args.config)
            logging.debug(' configuration: %s\n', config)

            generate(config)

        case 'mutate':
            return mutate(args.file)

        case 'continue':
            return continue_sequence(args.file)



from functools import reduce

from note import Note
from .types import Gene


def individual_to_melody(individual: list[Gene]) -> list[Note]:
    notes = []
    start = 0.0
    time = 0.0
    for i, gene in enumerate(individual):
        time += 0.5

        if gene.remaining_ticks > 1:
            continue

        notes.append(Note(int(gene.pitch), int(gene.velocity), start, time))
        start = time

    return notes


def melody_to_individual(melody: list[Note]) -> list[Gene]:
    genes = [Gene(0, 0, 1)] * 32
    for note in melody:
        ticks = int((note.end - note.start) * 2) + 1
        for i in range(ticks, 0, -1):
            genes[int(note.start * 2 + ticks - i - 1)] = Gene(note.pitch, note.velocity, i)

    return genes


def encodable(notes: list[Gene]) -> str:
    return reduce(lambda x, y: x + ',' + y, [str(note.pitch) for note in notes])



from enum import Enum, IntEnum
from dataclasses import dataclass
import logging


class Key(IntEnum):
    C = 0
    D = 1
    E = 2
    F = 3
    G = 4
    A = 5
    B = 6


class Mode(Enum):
    MAJOR = 1
    MINOR = 2
    PHRYGIAN = 3
    DORIAN = 4
    LYDIAN = 5
    MIXOLYDIAN = 6
    LOCRIAN = 7


@dataclass(frozen=True)
class Scale:
    notes: list[int]

    def __init__(self, key: Key, mode: Mode):
        def generate_notes(scale):
            return sorted([x + y if x + y < 127 else 0 for x in scale for y in range(key, 127, 12)])

        match mode:
            case Mode.MAJOR:
                object.__setattr__(self, 'notes', generate_notes([0, 2, 4, 5, 7, 9, 11]))
            case Mode.MINOR:
                object.__setattr__(self, 'notes', generate_notes([0, 2, 3, 5, 7, 8, 10]))
            case Mode.PHRYGIAN:
                object.__setattr__(self, 'notes', generate_notes([0, 1, 3, 5, 7, 8, 10]))
            case Mode.DORIAN:
                object.__setattr__(self, 'notes', generate_notes([0, 2, 3, 5, 7, 9, 10]))
            case Mode.LYDIAN:
                object.__setattr__(self, 'notes', generate_notes([0, 2, 4, 6, 7, 9, 11]))
            case Mode.MIXOLYDIAN:
                object.__setattr__(self, 'notes', generate_notes([0, 2, 4, 5, 7, 9, 10]))
            case Mode.LOCRIAN:
                object.__setattr__(self, 'notes', generate_notes([0, 1, 3, 5, 6, 8, 10]))


def match_key(key: str):
    match key:
        case 'c':
            return Key.C
        case 'd':
            return Key.D
        case 'e':
            return Key.E
        case 'f':
            return Key.F
        case 'g':
            return Key.G
        case 'a':
            return Key.A
        case 'b':
            return Key.B
    logging.error(f'Invalid key: {key}')


def match_mode(mode: str):
    match mode:
        case 'major':
            return Mode.MAJOR
        case 'minor':
            return Mode.MINOR
        case 'phrygian':
            return Mode.PHRYGIAN
        case 'dorian':
            return Mode.DORIAN
        case 'lydian':
            return Mode.LYDIAN
        case 'mixolydian':
            return Mode.MIXOLYDIAN
        case 'locrian':
            return Mode.LOCRIAN
    logging.error(f'Invalid mode: {mode}')


@dataclass
class Gene:
    pitch: int
    velocity: int
    remaining_ticks: int



from deap import creator, base

creator.create("Fitness", base.Fitness, weights=(-1.0, -1.0))
creator.create("KolmogorovFitness", base.Fitness, weights=(1.0,))

creator.create("Individual", list, fitness=creator.KolmogorovFitness)
creator.create("KolmogorovIndividual", list, fitness=creator.Fitness)



from io import TextIOWrapper

import configparser
import logging

from .config import Configuration
from .types import Scale, Mode, Key


def parse_config_for_generate(config_file: TextIOWrapper) -> Configuration:
    config = configparser.ConfigParser()
    config.read_file(config_file)

    try:
        return Configuration(
            bars=int(config['generate']['bars']),
            density=float(config['generate']['density']),
            rate=list(config['generate']['rate']),
            syncopation=float(config['generate']['syncopation']),
            scale=Scale(Key(config['generate']['scale']), Mode(config['generate']['mode']))
        )
    except KeyError as e:
        logging.error('Error: {}'.format(e))
        exit(1)


def parse_genetic_config():
    pass



from dataclasses import field, dataclass
from .types import Scale, Mode, Key


@dataclass
class Configuration:
    bars: int = 4
    notes_per_bar: int = 4
    ticks_per_note: int = 16
    density: float = 0.4
    rate: list[float] = field(default_factory=lambda: [1 / 4, 1 / 8])
    syncopation: float = 0.3
    scale: Scale = Scale(Key.C, Mode.MINOR)
    match: list[list[int]] = field(default=lambda: [])
    compression_method: str = 'LZ77'
    fitness_method: str = 'normal'
    consonance_rate: float = 0.8
    pitch_change_rate: float = 0.2
    length_change_rate: float = 0.1



from .syncopation import weighted_note_to_beat, density
from .conversion import individual_to_melody, encodable
from .compression import encode_lz77, encode_lz78, encode_lzw
from .config import Configuration
from .types import Gene

from itertools import takewhile
from deap import base, creator, algorithms, tools

import random
import logging


def fitness(genes: list[Gene]) -> tuple[float, float]:
    notes = individual_to_melody(genes)
    return abs(weighted_note_to_beat(notes) - 0.35), abs(density(genes) - 0.8)


def fitness_kolmogorov(config, genes: list[Gene]) -> tuple[float]:
    if config.compression_method == 'LZ77':
        compression = lambda x: len(encode_lz77(x))
    elif config.compression_method == 'LZ78':
        compression = lambda x: len(encode_lz78(x))
    elif config.compression_method == 'LZW':
        compression = lambda x: len(encode_lzw(x))
    else:
        raise AttributeError("compression method not found")

    ncd = lambda x, y: max(compression(x + y) - compression(x),
                           compression(y + x) - compression(y)) / max(compression(x), compression(y))

    distances = [ncd(encodable(genes), x) for x in config.match]

    return 1 / sum(distances),


def generator(config=Configuration()):
    prev_gene: Gene | None = None

    def create_random_gene() -> Gene:
        nonlocal prev_gene
        if prev_gene is not None and prev_gene.remaining_ticks > 1:
            prev_gene = Gene(prev_gene.pitch, prev_gene.velocity, prev_gene.remaining_ticks - 1)
            return prev_gene

        tick = min(config.rate)
        duration = int(random.choice(config.rate) / tick)
        key = random.choice(list(config.scale.notes)[30:40])
        velocity = random.randint(80, 100)
        prev_gene = Gene(key, velocity, duration)
        return prev_gene

    return create_random_gene


def mutation(config: Configuration, genes):
    for gene in genes:
        change = random.random()
        if change < config.pitch_change_rate:
            # change_2 = random.random()
            gene.pitch = random.choice(config.scale.notes[30:40])
            # if change_2 > config.consonance_rate:
            #     gene.pitch = random.choice(config.scale.notes[30:40])
            # else:
            #     gene.pitch = random.choice(list(set(POSSIBLE_NOTES) - set(config.scale.notes))[30:40])

        change = random.random()
        if change < config.length_change_rate:
            gene.remaining_ticks = 1
    return genes,


def check_remaining_ticks():
    def decorator(func):
        def wrapper(*args, **kwargs):
            offspring = func(*args, **kwargs)
            for genes in offspring:
                for i, gene in enumerate(genes):
                    if gene.remaining_ticks == 1:
                        continue
                    gene.remaining_ticks = len(list(takewhile(lambda x: x.pitch == gene.pitch, genes[i:])))
            return offspring

        return wrapper

    return decorator


def ea_simple_with_elitism(population, toolbox, cxpb, mutpb, ngen, hall_of_fame=None):
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    if hall_of_fame is None:
        raise ValueError("halloffame parameter must not be empty!")

    hall_of_fame.update(population)
    hof_size = len(hall_of_fame.items) if hall_of_fame.items else 0

    for gen in range(1, ngen + 1):
        offspring = toolbox.select(population, len(population) - hof_size)
        offspring = algorithms.varAnd(offspring, toolbox, cxpb, mutpb)

        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        offspring.extend(hall_of_fame.items)
        hall_of_fame.update(offspring)
        population[:] = offspring

    return population


def create_config(config=Configuration()) -> base.Toolbox:
    toolbox = base.Toolbox()
    toolbox.register("generator", generator())

    if config.fitness_method == 'kolmogorov':
        toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.generator, n=32)
        toolbox.register("evaluate", fitness_kolmogorov, config)
    else:
        toolbox.register("individual", tools.initRepeat, creator.KolmogorovIndividual, toolbox.generator, n=32)
        toolbox.register("evaluate", fitness)

    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("select", tools.selStochasticUniversalSampling)
    toolbox.register("mutate", mutation, config)
    toolbox.register("mate", tools.cxOnePoint)

    decorator = check_remaining_ticks()
    toolbox.decorate("mate", decorator)
    toolbox.decorate("mutate", decorator)

    return toolbox


def run_genetic_algorithm(config=Configuration()):
    toolbox = create_config(config)

    population = toolbox.population(100)
    hof = tools.HallOfFame(10)
    population = ea_simple_with_elitism(population, toolbox, cxpb=0.1, mutpb=0.1,
                                        ngen=100, hall_of_fame=hof)
    hof.update(population)
    best = hof.items[0]
    logging.info('-- Best Ever Individual = %s\n', best)
    logging.info('-- Best Ever Fitness -- %s\n', best.fitness.values)

    melody = individual_to_melody(best)
    return melody



from note import Note

from .conversion import encodable, melody_to_individual
from .genetic import individual_to_melody, create_config, run_genetic_algorithm
from .config import Configuration
from midiutil import MIDIFile

Sequence = list[Note]


def generate(config=Configuration()) -> Sequence:
    return run_genetic_algorithm(config)


def mutate(sequence: Sequence, config=Configuration()) -> Sequence:
    individual = melody_to_individual(sequence)
    toolbox = create_config(config)

    mutant = toolbox.clone(individual)
    ind, = toolbox.mutate(mutant)
    melody = individual_to_melody(ind)
    return melody


def continue_sequence(sequence: Sequence, config=Configuration()) -> Sequence:
    config.match = [encodable(melody_to_individual(sequence))]
    config.fitness_method = 'kolmogorov'

    return run_genetic_algorithm(config)


def combine(sequences: list[Sequence], config: Configuration = Configuration()) -> Sequence:
    config.match = [encodable(melody_to_individual(sequence)) for sequence in sequences]
    config.fitness_method = 'kolmogorov'
    return run_genetic_algorithm(config)


def write_file(notes: Sequence, path: str):
    midi = MIDIFile(1)
    midi.addTempo(0, 0, 120)

    for note in notes:
        midi.addNote(0, 0, note.pitch, note.start, note.end - note.start, note.velocity)

    with open(f'{path}', "wb") as output_file:
        midi.writeFile(output_file)



from note import Note
from math import floor


def weighted_note_to_beat(notes: list[Note]) -> float:
    if len(notes) < 5:
        return -1
    total = 0

    for note in notes:
        left = floor(note.start)
        right = left + 1
        distance = min(note.start - left, abs(note.start - right))

        if distance == 0:
            continue
        if right < note.end < right + 1:
            total += 2 / distance
        else:
            total += 1 / distance

    return total / len(notes)


def off_beatness(notes: list[Note]) -> float:
    b = [x for x in filter(lambda i: len(notes) % i == 0, range(1, len(notes) + 1))]
    weights = [0 if len(list(filter(lambda i: x % i == 0, b))) else 1 for x in notes]
    return sum(weights) / len(notes)


def density(notes) -> float:
    return len(list(filter(lambda note: note.velocity != 0, notes))) / len(notes)



import sys

import pytest

import env
import pybind11_cross_module_tests as cm
from pybind11_tests import exceptions as m


def test_std_exception(msg):
    with pytest.raises(RuntimeError) as excinfo:
        m.throw_std_exception()
    assert msg(excinfo.value) == "This exception was intentionally thrown."


def test_error_already_set(msg):
    with pytest.raises(RuntimeError) as excinfo:
        m.throw_already_set(False)
    assert (
        msg(excinfo.value)
        == "Internal error: pybind11::error_already_set called while Python error indicator not set."
    )

    with pytest.raises(ValueError) as excinfo:
        m.throw_already_set(True)
    assert msg(excinfo.value) == "foo"


def test_raise_from(msg):
    with pytest.raises(ValueError) as excinfo:
        m.raise_from()
    assert msg(excinfo.value) == "outer"
    assert msg(excinfo.value.__cause__) == "inner"


def test_raise_from_already_set(msg):
    with pytest.raises(ValueError) as excinfo:
        m.raise_from_already_set()
    assert msg(excinfo.value) == "outer"
    assert msg(excinfo.value.__cause__) == "inner"


def test_cross_module_exceptions(msg):
    with pytest.raises(RuntimeError) as excinfo:
        cm.raise_runtime_error()
    assert str(excinfo.value) == "My runtime error"

    with pytest.raises(ValueError) as excinfo:
        cm.raise_value_error()
    assert str(excinfo.value) == "My value error"

    with pytest.raises(ValueError) as excinfo:
        cm.throw_pybind_value_error()
    assert str(excinfo.value) == "pybind11 value error"

    with pytest.raises(TypeError) as excinfo:
        cm.throw_pybind_type_error()
    assert str(excinfo.value) == "pybind11 type error"

    with pytest.raises(StopIteration) as excinfo:
        cm.throw_stop_iteration()

    with pytest.raises(cm.LocalSimpleException) as excinfo:
        cm.throw_local_simple_error()
    assert msg(excinfo.value) == "external mod"

    with pytest.raises(KeyError) as excinfo:
        cm.throw_local_error()
    # KeyError is a repr of the key, so it has an extra set of quotes
    assert str(excinfo.value) == "'just local'"


# TODO: FIXME
@pytest.mark.xfail(
    "env.PYPY and env.MACOS",
    raises=RuntimeError,
    reason="Expected failure with PyPy and libc++ (Issue #2847 & PR #2999)",
)
def test_cross_module_exception_translator():
    with pytest.raises(KeyError):
        # translator registered in cross_module_tests
        m.throw_should_be_translated_to_key_error()


def test_python_call_in_catch():
    d = {}
    assert m.python_call_in_destructor(d) is True
    assert d["good"] is True


def ignore_pytest_unraisable_warning(f):
    unraisable = "PytestUnraisableExceptionWarning"
    if hasattr(pytest, unraisable):  # Python >= 3.8 and pytest >= 6
        dec = pytest.mark.filterwarnings(f"ignore::pytest.{unraisable}")
        return dec(f)
    else:
        return f


# TODO: find out why this fails on PyPy, https://foss.heptapod.net/pypy/pypy/-/issues/3583
@pytest.mark.xfail(env.PYPY, reason="Failure on PyPy 3.8 (7.3.7)", strict=False)
@ignore_pytest_unraisable_warning
def test_python_alreadyset_in_destructor(monkeypatch, capsys):
    hooked = False
    triggered = False

    if hasattr(sys, "unraisablehook"):  # Python 3.8+
        hooked = True
        # Don't take `sys.unraisablehook`, as that's overwritten by pytest
        default_hook = sys.__unraisablehook__

        def hook(unraisable_hook_args):
            exc_type, exc_value, exc_tb, err_msg, obj = unraisable_hook_args
            if obj == "already_set demo":
                nonlocal triggered
                triggered = True
            default_hook(unraisable_hook_args)
            return

        # Use monkeypatch so pytest can apply and remove the patch as appropriate
        monkeypatch.setattr(sys, "unraisablehook", hook)

    assert m.python_alreadyset_in_destructor("already_set demo") is True
    if hooked:
        assert triggered is True

    _, captured_stderr = capsys.readouterr()
    assert captured_stderr.startswith("Exception ignored in: 'already_set demo'")
    assert captured_stderr.rstrip().endswith("KeyError: 'bar'")


def test_exception_matches():
    assert m.exception_matches()
    assert m.exception_matches_base()
    assert m.modulenotfound_exception_matches_base()


def test_custom(msg):
    # Can we catch a MyException?
    with pytest.raises(m.MyException) as excinfo:
        m.throws1()
    assert msg(excinfo.value) == "this error should go to a custom type"

    # Can we translate to standard Python exceptions?
    with pytest.raises(RuntimeError) as excinfo:
        m.throws2()
    assert msg(excinfo.value) == "this error should go to a standard Python exception"

    # Can we handle unknown exceptions?
    with pytest.raises(RuntimeError) as excinfo:
        m.throws3()
    assert msg(excinfo.value) == "Caught an unknown exception!"

    # Can we delegate to another handler by rethrowing?
    with pytest.raises(m.MyException) as excinfo:
        m.throws4()
    assert msg(excinfo.value) == "this error is rethrown"

    # Can we fall-through to the default handler?
    with pytest.raises(RuntimeError) as excinfo:
        m.throws_logic_error()
    assert (
        msg(excinfo.value) == "this error should fall through to the standard handler"
    )

    # OverFlow error translation.
    with pytest.raises(OverflowError) as excinfo:
        m.throws_overflow_error()

    # Can we handle a helper-declared exception?
    with pytest.raises(m.MyException5) as excinfo:
        m.throws5()
    assert msg(excinfo.value) == "this is a helper-defined translated exception"

    # Exception subclassing:
    with pytest.raises(m.MyException5) as excinfo:
        m.throws5_1()
    assert msg(excinfo.value) == "MyException5 subclass"
    assert isinstance(excinfo.value, m.MyException5_1)

    with pytest.raises(m.MyException5_1) as excinfo:
        m.throws5_1()
    assert msg(excinfo.value) == "MyException5 subclass"

    with pytest.raises(m.MyException5) as excinfo:
        try:
            m.throws5()
        except m.MyException5_1 as err:
            raise RuntimeError("Exception error: caught child from parent") from err
    assert msg(excinfo.value) == "this is a helper-defined translated exception"


def test_nested_throws(capture):
    """Tests nested (e.g. C++ -> Python -> C++) exception handling"""

    def throw_myex():
        raise m.MyException("nested error")

    def throw_myex5():
        raise m.MyException5("nested error 5")

    # In the comments below, the exception is caught in the first step, thrown in the last step

    # C++ -> Python
    with capture:
        m.try_catch(m.MyException5, throw_myex5)
    assert str(capture).startswith("MyException5: nested error 5")

    # Python -> C++ -> Python
    with pytest.raises(m.MyException) as excinfo:
        m.try_catch(m.MyException5, throw_myex)
    assert str(excinfo.value) == "nested error"

    def pycatch(exctype, f, *args):
        try:
            f(*args)
        except m.MyException as e:
            print(e)

    # C++ -> Python -> C++ -> Python
    with capture:
        m.try_catch(
            m.MyException5,
            pycatch,
            m.MyException,
            m.try_catch,
            m.MyException,
            throw_myex5,
        )
    assert str(capture).startswith("MyException5: nested error 5")

    # C++ -> Python -> C++
    with capture:
        m.try_catch(m.MyException, pycatch, m.MyException5, m.throws4)
    assert capture == "this error is rethrown"

    # Python -> C++ -> Python -> C++
    with pytest.raises(m.MyException5) as excinfo:
        m.try_catch(m.MyException, pycatch, m.MyException, m.throws5)
    assert str(excinfo.value) == "this is a helper-defined translated exception"


def test_throw_nested_exception():
    with pytest.raises(RuntimeError) as excinfo:
        m.throw_nested_exception()
    assert str(excinfo.value) == "Outer Exception"
    assert str(excinfo.value.__cause__) == "Inner Exception"


# This can often happen if you wrap a pybind11 class in a Python wrapper
def test_invalid_repr():
    class MyRepr:
        def __repr__(self):
            raise AttributeError("Example error")

    with pytest.raises(TypeError):
        m.simple_bool_passthrough(MyRepr())


def test_local_translator(msg):
    """Tests that a local translator works and that the local translator from
    the cross module is not applied"""
    with pytest.raises(RuntimeError) as excinfo:
        m.throws6()
    assert msg(excinfo.value) == "MyException6 only handled in this module"

    with pytest.raises(RuntimeError) as excinfo:
        m.throws_local_error()
    assert not isinstance(excinfo.value, KeyError)
    assert msg(excinfo.value) == "never caught"

    with pytest.raises(Exception) as excinfo:
        m.throws_local_simple_error()
    assert not isinstance(excinfo.value, cm.LocalSimpleException)
    assert msg(excinfo.value) == "this mod"


class FlakyException(Exception):
    def __init__(self, failure_point):
        if failure_point == "failure_point_init":
            raise ValueError("triggered_failure_point_init")
        self.failure_point = failure_point

    def __str__(self):
        if self.failure_point == "failure_point_str":
            raise ValueError("triggered_failure_point_str")
        return "FlakyException.__str__"


@pytest.mark.parametrize(
    "exc_type, exc_value, expected_what",
    (
        (ValueError, "plain_str", "ValueError: plain_str"),
        (ValueError, ("tuple_elem",), "ValueError: tuple_elem"),
        (FlakyException, ("happy",), "FlakyException: FlakyException.__str__"),
    ),
)
def test_error_already_set_what_with_happy_exceptions(
    exc_type, exc_value, expected_what
):
    what, py_err_set_after_what = m.error_already_set_what(exc_type, exc_value)
    assert not py_err_set_after_what
    assert what == expected_what


@pytest.mark.skipif("env.PYPY", reason="PyErr_NormalizeException Segmentation fault")
def test_flaky_exception_failure_point_init():
    with pytest.raises(RuntimeError) as excinfo:
        m.error_already_set_what(FlakyException, ("failure_point_init",))
    lines = str(excinfo.value).splitlines()
    # PyErr_NormalizeException replaces the original FlakyException with ValueError:
    assert lines[:3] == [
        "pybind11::error_already_set: MISMATCH of original and normalized active exception types:"
        " ORIGINAL FlakyException REPLACED BY ValueError: triggered_failure_point_init",
        "",
        "At:",
    ]
    # Checking the first two lines of the traceback as formatted in error_string():
    assert "test_exceptions.py(" in lines[3]
    assert lines[3].endswith("): __init__")
    assert lines[4].endswith("): test_flaky_exception_failure_point_init")


def test_flaky_exception_failure_point_str():
    what, py_err_set_after_what = m.error_already_set_what(
        FlakyException, ("failure_point_str",)
    )
    assert not py_err_set_after_what
    lines = what.splitlines()
    if env.PYPY and len(lines) == 3:
        n = 3  # Traceback is missing.
    else:
        n = 5
    assert (
        lines[:n]
        == [
            "FlakyException: <MESSAGE UNAVAILABLE DUE TO ANOTHER EXCEPTION>",
            "",
            "MESSAGE UNAVAILABLE DUE TO EXCEPTION: ValueError: triggered_failure_point_str",
            "",
            "At:",
        ][:n]
    )


def test_cross_module_interleaved_error_already_set():
    with pytest.raises(RuntimeError) as excinfo:
        m.test_cross_module_interleaved_error_already_set()
    assert str(excinfo.value) in (
        "2nd error.",  # Almost all platforms.
        "RuntimeError: 2nd error.",  # Some PyPy builds (seen under macOS).
    )


def test_error_already_set_double_restore():
    m.test_error_already_set_double_restore(True)  # dry_run
    with pytest.raises(RuntimeError) as excinfo:
        m.test_error_already_set_double_restore(False)
    assert str(excinfo.value) == (
        "Internal error: pybind11::detail::error_fetch_and_normalize::restore()"
        " called a second time. ORIGINAL ERROR: ValueError: Random error."
    )


def test_pypy_oserror_normalization():
    # https://github.com/pybind/pybind11/issues/4075
    what = m.test_pypy_oserror_normalization()
    assert "this_filename_must_not_exist" in what



import note


def test_note_structure_exists():
    assert("Note" in dir(note))


def test_note_structure_has_attributes():
    assert("pitch" in dir(note.Note))
    assert("velocity" in dir(note.Note))
    assert("start" in dir(note.Note))
    assert("end" in dir(note.Note))


def test_can_create_note():
    test_note = note.Note(85, 100, 0, 1)
    assert(test_note.pitch == 85)
    assert(test_note.velocity == 100)
    assert(test_note.start == 0)
    assert(test_note.end == 1)




from midi_generator.commands import generate, mutate, continue_sequence, combine
from midi_generator.config import Configuration


def test_generate():
    notes = generate(Configuration())

    assert (len(notes) > 0)


def test_mutate():
    config = Configuration()
    notes = generate(config)
    mutant = mutate(notes)

    assert (len(mutant) > 0)


def test_continue():
    config = Configuration()
    notes = generate(config)
    continued = continue_sequence(notes, config)

    assert (len(continued) > 0)


def test_combine():
    config = Configuration()
    notes = [generate(config) for _ in range(4)]
    combined = combine(notes, config)

    assert (len(combined) > 0)



LOOP_WAIT = 1.0 / 200
BEATS_PER_BAR = 4
MIDI_EVENT_NOTE_ON = 0x90
NOTE_DURATIONS = [1, 1 / 2, 1 / 4, 1 / 8, 1 / 16]
DEFAULT_RATES = [1 / 2, 1 / 4, 1 / 8]
POSSIBLE_NOTES = [str(i) for i in range(128)]



from functools import reduce
from itertools import takewhile, zip_longest

from .constants import POSSIBLE_NOTES


def encode_lz77(string: str, window_size=100):
    encoded = string[: window_size + 1]
    i = window_size
    while i < len(string) - window_size:
        input_buffer = string[i: i + window_size + 1]
        window = string[i - window_size: i + window_size + 1]

        substring = max([reduce(lambda x, y: x + y,
                                map(lambda x: x[0],
                                    takewhile(lambda x: x[0] == x[1],
                                              zip_longest(string[i + j: i + window_size], input_buffer)
                                              )
                                    ),
                                '') for j in range(-window_size, 0)], key=len)

        if substring == '':
            i += 1
            encoded += f'0,0${string[i]}'
        else:
            i += len(substring)
            offset = window.find(substring)
            encoded += f'{str(offset)},{str(len(substring))}${string[i]}'

    return encoded


def encode_lz78(string: list):
    codes = dict()
    substring = ''
    index = 0
    encoded = ''

    for c in string:
        substring += c

        if substring not in codes:
            codes[substring] = index
            encoded += f'0,{c}' if substring == c else f'{str(codes[substring[:-1]])},{c}'
            index += 1
            substring = ''

    return encoded


def encode_lzw(string: list):
    codes = {c: i for i, c in enumerate(POSSIBLE_NOTES)}
    index = 4
    encoded = f'{codes[string[0]]},{codes[string[1]]},'
    substring = ''

    for c in string:
        substring += c

        if substring not in codes:
            index += 1
            codes[substring] = index
            encoded += str(codes[substring]) + ','
            substring = ''

    return encoded[:-1]
